# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

__all__: typing.Sequence[str] = [
    "ChannelConverter",
    "ColorConverter",
    "EmojiConverter",
    "GuildConverter",
    "InviteConverter",
    "MemberConverter",
    "PresenceConverter",
    "RoleConverter",
    "SnowflakeConverter",
    "UserConverter",
    "VoiceStateConverter",
]

import abc
import distutils.util
import inspect
import re
import typing
import warnings

from hikari import channels
from hikari import colors
from hikari import emojis
from hikari import guilds
from hikari import intents as intents_
from hikari import invites
from hikari import presences
from hikari import snowflakes
from hikari import users
from hikari import voices

from tanjun import errors
from tanjun import traits

_ValueT = typing.TypeVar("_ValueT", covariant=True)


class BaseConverter(abc.ABC, typing.Generic[_ValueT], traits.StatelessConverter[_ValueT]):
    __slots__: typing.Sequence[str] = ()
    __implementations: typing.MutableSet[typing.Type[BaseConverter[typing.Type[typing.Any]]]] = set()

    @classmethod
    def bind_client(cls, client: traits.Client, /) -> None:
        cache_bound = cls.cache_bound()
        if cache_bound and not client.cache_service:
            warnings.warn(
                f"Registered converter {cls!r} will always fail with a stateless client.", category=errors.StateWarning
            )
            return

        if cache_bound:
            required_intents = cls.intents()
            if (required_intents & client.shard_service.intents) != required_intents:
                warnings.warn(
                    f"Registered converter {cls.__name__!r} will not run as expected "
                    f"when {required_intents!r} intent(s) are not declared",
                    category=errors.StateWarning,
                )

    @classmethod
    def bind_component(self, _: traits.Component, /) -> None:
        pass

    @classmethod
    def get_from_type(
        cls, type_: typing.Type[_ValueT]
    ) -> typing.Optional[typing.Type[BaseConverter[typing.Type[_ValueT]]]]:
        for converter in cls.__implementations:
            is_inheritable = converter.is_inheritable()
            if is_inheritable and issubclass(type_, converter.types()):
                return converter

            if not is_inheritable and type_ in converter.types():
                return converter

        return None

    @classmethod
    def implementations(cls) -> typing.MutableSet[typing.Type[BaseConverter[typing.Type[typing.Any]]]]:
        return cls.__implementations

    @classmethod
    @abc.abstractmethod
    def cache_bound(cls) -> bool:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    async def convert(cls, ctx: traits.Context, argument: str, /) -> _ValueT:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def intents(cls) -> intents_.Intents:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def is_inheritable(cls) -> bool:
        raise NotImplementedError

    @classmethod
    @abc.abstractmethod
    def types(cls) -> typing.Tuple[typing.Type[typing.Any], ...]:
        raise NotImplementedError


class ChannelConverter(BaseConverter[channels.GuildChannel]):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    def cache_bound(cls) -> bool:
        return True

    @classmethod
    async def convert(cls, ctx: traits.Context, argument: str, /) -> channels.GuildChannel:
        if ctx.client.cache_service:
            channel_id = ChannelIDParser.match_id(argument, message="No valid channel mention or ID  found")
            if channel := ctx.client.cache_service.cache.get_guild_channel(channel_id):
                return channel

        raise ValueError("Couldn't find channel")

    @classmethod
    def intents(cls) -> intents_.Intents:
        return intents_.Intents.GUILDS

    @classmethod
    def is_inheritable(cls) -> bool:
        return True

    @classmethod
    def types(cls) -> typing.Tuple[typing.Type[typing.Any], ...]:
        return (channels.GuildChannel,)


class ColorConverter(BaseConverter[colors.Color]):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    def cache_bound(cls) -> bool:
        return False

    @classmethod
    async def convert(cls, _: traits.Context, argument: str, /) -> typing.Any:
        values = argument.split(" ")
        if all(value.isdigit() for value in values):
            return colors.Color.of(*map(int, values))

        return colors.Color.of(*values)

    @classmethod
    def intents(cls) -> intents_.Intents:
        return intents_.Intents.NONE

    @classmethod
    def is_inheritable(cls) -> bool:
        return False

    @classmethod
    def types(cls) -> typing.Tuple[typing.Type[typing.Any], ...]:
        return (colors.Color,)


class EmojiConverter(BaseConverter[emojis.KnownCustomEmoji]):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    def cache_bound(cls) -> bool:
        return True

    @classmethod
    async def convert(cls, ctx: traits.Context, argument: str, /) -> emojis.KnownCustomEmoji:
        if ctx.client.cache_service:
            emoji_id = EmojiIDParser.match_id(argument, message="No valid emoji or emoji ID found")
            if emoji := ctx.client.cache_service.cache.get_emoji(emoji_id):
                return emoji

        raise ValueError("Couldn't find emoji")

    @classmethod
    def intents(cls) -> intents_.Intents:
        return intents_.Intents.GUILD_EMOJIS

    @classmethod
    def is_inheritable(cls) -> bool:
        return True

    @classmethod
    def types(cls) -> typing.Tuple[typing.Type[typing.Any], ...]:
        return (emojis.CustomEmoji,)


class GuildConverter(BaseConverter[guilds.GatewayGuild]):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    def cache_bound(cls) -> bool:
        return True

    @classmethod
    async def convert(cls, ctx: traits.Context, argument: str, /) -> guilds.GatewayGuild:
        if ctx.client.cache_service:
            guild_id = SnowflakeParser.match_id(argument, message="No valid guild ID found")
            if guild := ctx.client.cache_service.cache.get_guild(guild_id):
                return guild

        raise ValueError("Couldn't find guild")

    @classmethod
    def intents(cls) -> intents_.Intents:
        return intents_.Intents.GUILDS

    @classmethod
    def is_inheritable(cls) -> bool:
        return True

    @classmethod
    def types(cls) -> typing.Tuple[typing.Type[typing.Any], ...]:
        return (guilds.Guild,)


class InviteConverter(BaseConverter[invites.InviteWithMetadata]):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    def cache_bound(cls) -> bool:
        return True

    @classmethod
    async def convert(cls, ctx: traits.Context, argument: str, /) -> invites.InviteWithMetadata:
        if ctx.client.cache_service:
            if invite := ctx.client.cache_service.cache.get_invite(argument):
                return invite

        raise ValueError("Couldn't find invite")

    @classmethod
    def intents(cls) -> intents_.Intents:
        return intents_.Intents.GUILD_INVITES

    @classmethod
    def is_inheritable(cls) -> bool:
        return True

    @classmethod
    def types(cls) -> typing.Tuple[typing.Type[typing.Any], ...]:
        return (invites.Invite,)


class MemberConverter(BaseConverter[guilds.Member]):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    def cache_bound(cls) -> bool:
        return True

    @classmethod
    async def convert(cls, ctx: traits.Context, argument: str, /) -> guilds.Member:
        if ctx.message.guild_id is None:
            raise ValueError("Cannot get a member from a DM channel")

        if ctx.client.cache_service:
            member_id = UserIDParser.match_id(argument, message="No valid user mention or ID found")
            if member := ctx.client.cache_service.cache.get_member(ctx.message.guild_id, member_id):
                return member

        raise ValueError("Couldn't find member in this guild")

    @classmethod
    def intents(cls) -> intents_.Intents:
        return intents_.Intents.GUILD_MEMBERS

    @classmethod
    def is_inheritable(cls) -> bool:
        return False

    @classmethod
    def types(cls) -> typing.Tuple[typing.Type[typing.Any], ...]:
        return (guilds.Member,)


class PresenceConverter(BaseConverter[presences.MemberPresence]):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    def cache_bound(cls) -> bool:
        return True

    @classmethod
    async def convert(cls, ctx: traits.Context, argument: str, /) -> presences.MemberPresence:
        if ctx.message.guild_id is None:
            raise ValueError("Cannot get a presence from a DM channel")

        if ctx.client.cache_service:
            user_id = UserIDParser.match_id(argument, message="No valid member mention or ID  found")
            if user := ctx.client.cache_service.cache.get_presence(ctx.message.guild_id, user_id):
                return user

        raise ValueError("Couldn't find presence in current guild")


class RoleConverter(BaseConverter[guilds.Role]):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    def cache_bound(cls) -> bool:
        return True

    @classmethod
    async def convert(cls, ctx: traits.Context, argument: str, /) -> guilds.Role:
        if ctx.client.cache_service:
            role_id = SnowflakeParser.match_id(argument, message="No valid role mention or ID  found")
            if role := ctx.client.cache_service.cache.get_role(role_id):
                return role

        raise ValueError("Couldn't find role")

    @classmethod
    def intents(cls) -> intents_.Intents:
        return intents_.Intents.GUILDS

    @classmethod
    def is_inheritable(cls) -> bool:
        return False

    @classmethod
    def types(cls) -> typing.Tuple[typing.Type[typing.Any], ...]:
        return (guilds.Role,)


class BaseSnowflakeParser:
    __slots__: typing.Sequence[str] = ()

    @classmethod
    @abc.abstractmethod
    def regex(cls) -> typing.Pattern[str]:
        raise NotImplementedError

    @classmethod
    def match_id(cls, value: str, *, message: str = "No valid mention or ID found") -> snowflakes.Snowflake:
        result: typing.Optional[snowflakes.Snowflake] = None
        value = value.strip()
        if value.isdigit():
            result = snowflakes.Snowflake(value)

        else:
            try:
                result = snowflakes.Snowflake(next(cls.regex().finditer(value)).groups()[0])

            except StopIteration:
                pass

        # We should also range check the provided ID.
        if result is not None and snowflakes.Snowflake.min() <= result <= snowflakes.Snowflake.max():
            return result

        raise ValueError(message) from None


class SnowflakeParser(BaseSnowflakeParser):
    __slots__: typing.Sequence[str] = ()
    _pattern = re.compile(r"<[@&?!#a]{0,3}(?::\w+:)?(\d+)>")

    @classmethod
    def regex(cls) -> typing.Pattern[str]:
        return cls._pattern


class ChannelIDParser(BaseSnowflakeParser):
    __slots__: typing.Sequence[str] = ()
    _pattern = re.compile(r"<#(\d+)>")

    @classmethod
    def regex(cls) -> typing.Pattern[str]:
        return cls._pattern


class EmojiIDParser(BaseSnowflakeParser):
    __slots__: typing.Sequence[str] = ()
    _pattern = re.compile(r"<a?:\w+:(\d+)>")

    @classmethod
    def regex(cls) -> typing.Pattern[str]:
        return cls._pattern


class UserIDParser(BaseSnowflakeParser):
    __slots__: typing.Sequence[str] = ()
    _pattern = re.compile(r"<@!?(\d+)>")

    @classmethod
    def regex(cls) -> typing.Pattern[str]:
        return cls._pattern


class SnowflakeConverter(BaseConverter[snowflakes.Snowflake]):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    def cache_bound(cls) -> bool:
        return False

    @classmethod
    async def convert(cls, _: traits.Context, argument: str, /) -> snowflakes.Snowflake:
        return SnowflakeParser.match_id(argument, message="No valid ID found")

    @classmethod
    def intents(cls) -> intents_.Intents:
        return intents_.Intents.NONE

    @classmethod
    def is_inheritable(cls) -> bool:
        return False

    @classmethod
    def types(cls) -> typing.Tuple[typing.Type[typing.Any], ...]:
        return (snowflakes.Snowflake,)


class UserConverter(BaseConverter[users.User]):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    def cache_bound(cls) -> bool:
        return True

    @classmethod
    async def convert(cls, ctx: traits.Context, argument: str, /) -> users.User:
        if ctx.client.cache_service:
            user_id = UserIDParser.match_id(argument, message="No valid user mention or ID  found")
            if user := ctx.client.cache_service.cache.get_user(user_id):
                return user

        raise ValueError("Couldn't find user")

    @classmethod
    def intents(cls) -> intents_.Intents:
        return intents_.Intents.GUILD_MEMBERS

    @classmethod
    def is_inheritable(cls) -> bool:
        return False

    @classmethod
    def types(cls) -> typing.Tuple[typing.Type[typing.Any], ...]:
        return (users.User,)


class VoiceStateConverter(BaseConverter[voices.VoiceState]):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    def cache_bound(cls) -> bool:
        return True

    @classmethod
    async def convert(cls, ctx: traits.Context, argument: str, /) -> voices.VoiceState:
        if ctx.message.guild_id is None:
            raise ValueError("Cannot get a voice state from a DM channel")

        if ctx.client.cache_service:
            user_id = UserIDParser.match_id(argument, message="No valid user mention or ID  found")
            if user := ctx.client.cache_service.cache.get_voice_state(ctx.message.guild_id, user_id):
                return user

        raise ValueError("Voice state couldn't be found for current guild")

    @classmethod
    def intents(cls) -> intents_.Intents:
        return intents_.Intents.GUILD_VOICE_STATES

    @classmethod
    def is_inheritable(cls) -> bool:
        return False

    @classmethod
    def types(cls) -> typing.Tuple[typing.Type[typing.Any], ...]:
        return (voices.VoiceState,)


for _cls in vars().copy().values():
    if inspect.isclass(_cls) and issubclass(_cls, BaseConverter):
        BaseConverter.implementations().add(_cls)

del _cls


_BUILTIN_TYPE_OVERRIDES: typing.Mapping[typing.Callable[..., typing.Any], typing.Callable[[str], typing.Any]] = {
    bool: distutils.util.strtobool,
    bytes: lambda d: bytes(d, "utf-8"),
    bytearray: lambda d: bytearray(d, "utf-8"),
}


def override_builtin_type(cls: traits.ConverterT) -> traits.ConverterT:
    if callable(cls):
        return _BUILTIN_TYPE_OVERRIDES.get(cls, cls)

    return cls



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Collection of utility functions used within Tanjun."""
from __future__ import annotations

__all__: typing.Sequence[str] = [
    "async_chain",
    "await_if_async",
    "gather_checks",
    "ALL_PERMISSIONS",
    "calculate_permissions",
    "with_function_wrapping",
]

import asyncio
import types
import typing

from hikari import channels
from hikari import errors as hikari_errors
from hikari import permissions as permissions_
from yuyo import backoff

from tanjun import errors as tanjun_errors

if typing.TYPE_CHECKING:
    from hikari import guilds
    from hikari import snowflakes

    from tanjun import traits as tanjun_traits


_ResourceT = typing.TypeVar("_ResourceT")
_ValueT = typing.TypeVar("_ValueT")


async def async_chain(iterable: typing.Iterable[typing.AsyncIterable[_ValueT]]) -> typing.AsyncIterator[_ValueT]:
    """Make an asynchronous iterator of the elements within multiple asynchronous iterators."""
    for async_iterable in iterable:
        async for value in async_iterable:
            yield value


async def await_if_async(
    callback: typing.Callable[..., typing.Union[_ValueT, typing.Awaitable[_ValueT]]], *args: typing.Any
) -> _ValueT:
    """Resole any awaitable returned by a function call.

    Parameters
    ----------
    callback : typing.Callable[..., typing.Union[_ValueT, typing.Awaitable[_ValueT]]
        The async or non-async function to call.

    Other Parameters
    ----------------
    *args : typing.Any
        A variable amount of positional arguments to pass through when calling
        `callback`.

    Returns
    -------
    _ValueT
        The resolved result of the passed function.
    """
    result = callback(*args)

    if isinstance(result, typing.Awaitable):
        # For some reason MYPY thinks this returns typing.Any
        return typing.cast(_ValueT, await result)

    return result


async def _wrap_check(check: typing.Awaitable[bool]) -> bool:
    # We raise on `False` to let asyncio.gather stop execution on the first failed
    # check rather than waiting for all checks to fail.
    if not await check:
        raise tanjun_errors.FailedCheck()

    return True


async def gather_checks(checks: typing.Iterable[typing.Awaitable[bool]]) -> bool:
    """Gather a collection of checks.

    Parameters
    ----------
    checks : typing.Iterable[typing.Awaitable[bool]]
        An iterable of check awaitables which each return `builtin.bool`.
        These may raise `hikari.errors.FailedCheck` as an alternative to
        returning `False`.

    Returns
    -------
    bool
        Whether all the checks passed or not.
    """
    try:
        await asyncio.gather(*map(_wrap_check, checks))
        # _wrap_check will raise FailedCheck if a false is received so if we get
        # this far then it's True.
        return True

    except tanjun_errors.FailedCheck:
        return False


async def fetch_resource(
    retry: backoff.Backoff, call: typing.Callable[..., typing.Awaitable[_ResourceT]], *args: typing.Any
) -> _ResourceT:
    """A utility function for retrying a request used by Tanjun internally."""
    retry.reset()
    async for _ in retry:
        try:
            return await call(*args)

        except hikari_errors.RateLimitedError as exc:
            if exc.retry_after > 5:
                raise

            retry.set_next_backoff(exc.retry_after)

        except hikari_errors.InternalServerError:
            continue

    else:
        return await call(*args)


ALL_PERMISSIONS = permissions_.Permissions.NONE
"""All of the known permissions based on the linked version of Hikari."""

for _permission in permissions_.Permissions:
    ALL_PERMISSIONS |= _permission

del _permission


def _calculate_channel_overwrites(
    channel: channels.GuildChannel, member: guilds.Member, permissions: permissions_.Permissions
) -> permissions_.Permissions:
    if everyone_overwrite := channel.permission_overwrites.get(member.guild_id):
        permissions &= ~everyone_overwrite.deny
        permissions |= everyone_overwrite.allow

    deny = permissions_.Permissions.NONE
    allow = permissions_.Permissions.NONE

    for overwrite in filter(None, map(channel.permission_overwrites.get, member.role_ids)):
        deny |= overwrite.deny
        allow |= overwrite.allow

    permissions &= ~deny
    permissions |= allow

    if member_overwrite := channel.permission_overwrites.get(member.user.id):
        permissions &= ~member_overwrite.deny
        permissions |= member_overwrite.allow

    return permissions


def _calculate_role_permissions(
    roles: typing.Mapping[snowflakes.Snowflake, guilds.Role], member: guilds.Member
) -> permissions_.Permissions:
    permissions = roles[member.guild_id].permissions

    for role in filter(None, map(roles.get, member.role_ids)):
        if role.id != member.guild_id:
            permissions |= role.permissions

    return permissions


async def calculate_permissions(
    client: tanjun_traits.Client,
    member: guilds.Member,
    /,
    *,
    channel: typing.Optional[snowflakes.SnowflakeishOr[channels.GuildChannel]],
) -> permissions_.Permissions:
    """Calculate the permissions a member has within a guild.

    Parameters
    ----------
    client : tanjun.traits.Client
        The Tanjun client to use for lookups.
    member : hikari.guilds.Member
        The object of the member to calculate the permissions for.
    channel : typing.Optional[hikari.snowflakes.SnowflakeishOr[hikari.channels.GuildChannel]]
        The object of ID of the channel to get their permissions in.
        If left as `builtins.None` then this will return their base guild
        permissions.

    !!! note
        This function will fallback to REST requests if cache lookups fail or
        are not possible.

    Returns
    -------
    hikari.permissions.Permissions
        The calculated permissions.
    """
    # The ordering of how this adds and removes permissions does matter.
    # For more information see https://discord.com/developers/docs/topics/permissions#permission-hierarchy.
    retry = backoff.Backoff(maximum=5, max_retries=4)
    found_channel = channel if isinstance(channel, channels.GuildChannel) else None
    guild: typing.Optional[guilds.Guild] = None
    roles: typing.Optional[typing.Mapping[snowflakes.Snowflake, guilds.Role]] = None
    if client.cache_service:
        if not found_channel and channel:
            found_channel = client.cache_service.cache.get_guild_channel(snowflakes.Snowflake(channel))

        guild = client.cache_service.cache.get_guild(member.guild_id)
        roles = client.cache_service.cache.get_roles_view_for_guild(member.guild_id)

    if not guild:
        raw_guild = await fetch_resource(retry, client.rest_service.rest.fetch_guild, member.guild_id)
        guild = raw_guild

    # Guild owners are implicitly admins.
    if guild.owner_id == member.user.id:
        return ALL_PERMISSIONS

    if not roles:
        raw_roles = await fetch_resource(retry, client.rest_service.rest.fetch_roles, member.guild_id)
        roles = {role.id: role for role in raw_roles}

    # Admin permission overrides all overwrites and is only applicable to roles.
    if (permissions := _calculate_role_permissions(roles, member)) & permissions.ADMINISTRATOR:
        return ALL_PERMISSIONS

    if not channel:
        return permissions

    if not found_channel:
        raw_channel = await fetch_resource(retry, client.rest_service.rest.fetch_channel, channel)
        assert isinstance(raw_channel, channels.GuildChannel), "We shouldn't get DM channels in guilds."
        found_channel = raw_channel

    return _calculate_channel_overwrites(found_channel, member, permissions)


def with_function_wrapping(obj: typing.Any, function_field: str, /) -> None:
    """Utility function for making an object wrap a function at runtime.

    Parameters
    ----------
    obj : typing.Any
        The initialised object to update the wrapping information for.
    function_field : str
        The name of the attribute which is being wrapped. This should point to a
        function.
    """
    obj.__annotations__ = getattr(obj, function_field).__annotations__
    # For the sake of presenting as the right signature the objects this will be applied to should pre-define
    # their own __call__ method which assumes "self" is already bound to the contained function.
    obj.__call__ = types.MethodType(lambda self, *args, **kwargs: getattr(self, function_field)(*args, **kwargs), obj)
    obj.__doc__ = getattr(obj, function_field).__doc__
    obj.__module__ = getattr(obj, function_field).__module__
    obj.__wrapped__ = getattr(obj, function_field)



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

__all__: typing.Sequence[str] = ["Context"]

import typing

from hikari import snowflakes

from tanjun import traits

if typing.TYPE_CHECKING:
    from hikari import messages
    from hikari import traits as hikari_traits
    from hikari.api import shard as shard_


class Context(traits.Context):
    """Standard implementation of a command context as used within Tanjun."""

    __slots__: typing.Sequence[str] = (
        "_client",
        "command",
        "content",
        "_message",
        "_rest",
        "triggering_name",
        "triggering_prefix",
        "_shard",
    )

    def __init__(
        self,
        client: traits.Client,
        /,
        content: str,
        message: messages.Message,
        *,
        command: typing.Optional[traits.ExecutableCommand] = None,
        triggering_name: typing.Optional[str] = None,
        triggering_prefix: typing.Optional[str] = None,
    ) -> None:
        if message.content is None:
            raise ValueError("Cannot spawn context with a contentless message.")

        self._client = client
        self.command = command
        self.content = content
        self._message = message
        self.triggering_name = triggering_name
        self.triggering_prefix = triggering_prefix

    def __repr__(self) -> str:
        return f"Context <{self.message!r}, {self.command!r}>"

    @property
    def cache_service(self) -> typing.Optional[hikari_traits.CacheAware]:
        return self._client.cache_service

    @property
    def client(self) -> traits.Client:
        return self._client

    @property
    def dispatch_service(self) -> hikari_traits.DispatcherAware:
        return self._client.dispatch_service

    @property
    def message(self) -> messages.Message:
        return self._message

    @property
    def rest_service(self) -> hikari_traits.RESTAware:
        return self._client.rest_service

    @property
    def shard_service(self) -> hikari_traits.ShardAware:
        return self._client.shard_service

    @property
    def shard(self) -> shard_.GatewayShard:
        if self.message.guild_id is not None:
            shard_id = snowflakes.calculate_shard_id(self.shard_service, self.message.guild_id)

        else:
            shard_id = 0

        return self.shard_service.shards[shard_id]



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

__all__: typing.Sequence[str] = [
    # checks.py
    "checks",
    "with_dm_check",
    "with_guild_check",
    "with_nsfw_check",
    "with_sfw_check",
    "with_owner_check",
    "with_author_permission_check",
    "with_bot_permission_check",
    # clients.py
    "clients",
    "as_loader",
    "Client",
    # command.py
    "commands",
    "Command",
    "CommandGroup",
    # components.py
    "components",
    "as_command",
    "as_event",
    "as_group",
    "Component",
    # context.py
    "context",
    "Context",
    # conversion.py
    "conversion",
    "ChannelConverter",
    "ColorConverter",
    "EmojiConverter",
    "GuildConverter",
    "InviteConverter",
    "MemberConverter",
    "PresenceConverter",
    "RoleConverter",
    "SnowflakeConverter",
    "UserConverter",
    "VoiceStateConverter",
    # errors.py
    "errors",
    "CommandError",
    "ConversionError",
    "FailedCheck",
    "NotEnoughArgumentsError",
    "TooManyArgumentsError",
    "ParserError",
    "TanjunError",
    "TanjunWarning",
    "StateWarning",
    # hooks.py
    "hooks",
    "Hooks",
    # parsing.py
    "parsing",
    "Argument",
    "Option",
    "ShlexParser",
    "parser_descriptor",
    "verify_parameters",
    "with_argument",
    "with_greedy_argument",
    "with_multi_argument",
    "with_option",
    "with_multi_option",
    "with_parser",
    "with_typed_parameters",
    # traits.py
    "traits",
]

import typing

from tanjun import traits
from tanjun.checks import *
from tanjun.clients import *
from tanjun.commands import *
from tanjun.components import *
from tanjun.context import *
from tanjun.conversion import *
from tanjun.errors import *
from tanjun.hooks import *
from tanjun.parsing import *



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

__all__: typing.Sequence[str] = ["Command", "CommandGroup"]

import typing

from hikari import errors as hikari_errors
from yuyo import backoff

from tanjun import errors
from tanjun import hooks as hooks_
from tanjun import traits
from tanjun import utilities


class FoundCommand(traits.FoundCommand):
    __slots__: typing.Sequence[str] = ("command", "name", "prefix")

    def __init__(self, command_: traits.ExecutableCommand, name: str, /) -> None:
        self.command = command_
        self.name = name


class Command(traits.ExecutableCommand):
    __slots__: typing.Sequence[str] = (
        "_checks",
        "_component",
        "_function",
        "hooks",
        "_metadata",
        "_names",
        "parent",
        "parser",
    )

    def __init__(
        self,
        function: traits.CommandFunctionT,
        name: str,
        /,
        *names: str,
        checks: typing.Optional[typing.Iterable[traits.CheckT]] = None,
        hooks: typing.Optional[traits.Hooks] = None,
        metadata: typing.Optional[typing.MutableMapping[typing.Any, typing.Any]] = None,
        parser: typing.Optional[traits.Parser] = None,
    ) -> None:
        self._checks = set(checks) if checks else set()
        self._component: typing.Optional[traits.Component] = None
        self._function = function
        self.hooks: traits.Hooks = hooks or hooks_.Hooks()
        self._metadata = metadata or {}
        self._names = {name, *names}
        self.parent: typing.Optional[traits.ExecutableCommandGroup] = None
        self.parser = parser

    def __repr__(self) -> str:
        return f"Command <{self._names}>"

    @property
    def checks(self) -> typing.AbstractSet[traits.CheckT]:
        return frozenset(self._checks)

    @property
    def component(self) -> typing.Optional[traits.Component]:
        return self._component

    @property
    def function(self) -> traits.CommandFunctionT:
        return self._function

    @property
    def metadata(self) -> typing.MutableMapping[typing.Any, typing.Any]:
        return self._metadata

    @property
    def names(self) -> typing.AbstractSet[str]:
        return frozenset(self._names)

    def add_check(self, check: traits.CheckT, /) -> None:
        self._checks.add(check)

    def remove_check(self, check: traits.CheckT, /) -> None:
        self._checks.remove(check)

    def with_check(self, check: traits.CheckT, /) -> traits.CheckT:
        self.add_check(check)
        return check

    async def check_context(
        self, ctx: traits.Context, /, *, name_prefix: str = ""
    ) -> typing.AsyncIterator[traits.FoundCommand]:
        if found := next(self.check_name(ctx.content[len(name_prefix) :].lstrip()), None):
            if await utilities.gather_checks(utilities.await_if_async(check, ctx) for check in self._checks):
                yield found

    def add_name(self, name: str, /) -> None:
        self._names.add(name)

    def check_name(self, name: str, /) -> typing.Iterator[traits.FoundCommand]:
        for own_name in self._names:
            # Here we enforce that a name must either be at the end of content or be followed by a space. This helps
            # avoid issues with ambiguous naming where a command with the names "name" and "names" may sometimes hit
            # the former before the latter when triggered with the latter, leading to the command potentially being
            # inconsistently parsed.
            if name.startswith(own_name) and (name == own_name or name[len(own_name)] == " "):
                yield FoundCommand(self, own_name)
                break

    def remove_name(self, name: str, /) -> None:
        self._names.remove(name)

    def bind_client(self, client: traits.Client, /) -> None:
        if self.parser:
            self.parser.bind_client(client)

    def bind_component(self, component: traits.Component, /) -> None:
        pass

    async def execute(
        self, ctx: traits.Context, /, *, hooks: typing.Optional[typing.MutableSet[traits.Hooks]] = None
    ) -> bool:
        try:
            if await self.hooks.trigger_pre_execution(ctx, hooks=hooks) is False:
                return True

            if self.parser is not None:
                args, kwargs = await self.parser.parse(ctx)

            else:
                args = []
                kwargs = {}

            await self._function(ctx, *args, **kwargs)

        except errors.CommandError as exc:
            if not exc.message:
                return True

            response = exc.message if len(exc.message) <= 2000 else exc.message[:1997] + "..."
            retry = backoff.Backoff(max_retries=5, maximum=2)
            # TODO: preemptive cache based permission checks before throwing to the REST gods.
            async for _ in retry:
                try:
                    await ctx.message.reply(content=response)

                except hikari_errors.RateLimitedError as retry_error:
                    if retry_error.retry_after > 4:
                        raise

                    retry.set_next_backoff(retry_error.retry_after)  # TODO: check if this is too large.

                except hikari_errors.InternalServerError:
                    continue

                except (hikari_errors.ForbiddenError, hikari_errors.NotFoundError):
                    break

                else:
                    break

        except errors.ParserError as exc:
            await self.hooks.trigger_parser_error(ctx, exc, hooks=hooks)

        except Exception as exc:
            await self.hooks.trigger_error(ctx, exc, hooks=hooks)
            raise

        else:
            # TODO: how should this be handled around CommandError?
            await self.hooks.trigger_success(ctx, hooks=hooks)

        finally:
            await self.hooks.trigger_post_execution(ctx, hooks=hooks)

        return True


class CommandGroup(Command, traits.ExecutableCommandGroup):
    __slots__: typing.Sequence[str] = ("_commands",)

    def __init__(
        self,
        function: traits.CommandFunctionT,
        name: str,
        /,
        *names: str,
        checks: typing.Optional[typing.Iterable[traits.CheckT]] = None,
        hooks: typing.Optional[traits.Hooks] = None,
        metadata: typing.Optional[typing.MutableMapping[typing.Any, typing.Any]] = None,
        parser: typing.Optional[traits.Parser] = None,
    ) -> None:
        super().__init__(function, name, *names, checks=checks, hooks=hooks, metadata=metadata, parser=parser)
        self._commands: typing.MutableSet[traits.ExecutableCommand] = set()

    def __repr__(self) -> str:
        return f"CommandGroup <{len(self._commands)}: {self._names}>"

    @property
    def commands(self) -> typing.AbstractSet[traits.ExecutableCommand]:
        return frozenset(self._commands)

    def add_command(self, command: traits.ExecutableCommand, /) -> None:
        command.parent = self
        self._commands.add(command)

    def remove_command(self, command: traits.ExecutableCommand, /) -> None:
        command.parent = None
        self._commands.remove(command)

    def with_command(
        self,
        name: str,
        /,
        *names: str,
        checks: typing.Optional[typing.Iterable[traits.CheckT]] = None,
        hooks: typing.Optional[traits.Hooks] = None,
        parser: typing.Optional[traits.Parser] = None,
    ) -> typing.Callable[[traits.CommandFunctionT], traits.CommandFunctionT]:
        def decorator(function: traits.CommandFunctionT, /) -> traits.CommandFunctionT:
            self.add_command(Command(function, name, *names, checks=checks, hooks=hooks, parser=parser))
            return function

        return decorator

    def bind_client(self, client: traits.Client, /) -> None:
        super().bind_client(client)
        for command in self._commands:
            command.bind_client(client)

    # I sure hope this plays well with command group recursion cause I am waaaaaaaaaaaaaay too lazy to test that myself.
    async def execute(
        self, ctx: traits.Context, /, *, hooks: typing.Optional[typing.MutableSet[traits.Hooks]] = None
    ) -> bool:
        if ctx.message.content is None:
            raise ValueError("Cannot execute a command with a contentless message")

        if self.hooks and hooks:
            hooks.add(self.hooks)

        elif self.hooks:
            hooks = {self.hooks}

        for command in self._commands:
            async for result in command.check_context(ctx):
                # triggering_prefix should never be None here but for the sake of covering all cases if it is then we
                # assume an empty string.
                # If triggering_name is None then we assume an empty string for that as well.
                content = ctx.message.content.lstrip()[len(ctx.triggering_prefix or "") :].lstrip()[
                    len(ctx.triggering_name or "") :
                ]
                space_len = len(content) - len(content.lstrip())
                ctx.triggering_name = (ctx.triggering_name or "") + (" " * space_len) + result.name
                ctx.content = ctx.content[space_len + len(result.name) :].lstrip()
                await result.command.execute(ctx, hooks=hooks)
                return True

        return await super().execute(ctx, hooks=hooks)



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""Interfaces of the objects used within Tanjun."""
from __future__ import annotations

__all__: typing.Sequence[str] = [
    "CommandDescriptor",
    "ListenerDescriptor",
    "LoadableDescriptor",
    "ParserDescriptor",
    "ConverterT",
    "ParserHookT",
    "ErrorHookT",
    "HookT",
    "PreExecutionHookT",
    "CheckT",
    "ValueT",
    "Context",
    "Converter",
    "StatelessConverter",
    "Hooks",
    "Executable",
    "FoundCommand",
    "ExecutableCommand",
    "ExecutableCommandGroup",
    "Component",
    "Client",
    "UNDEFINED_DEFAULT",
    "Parameter",
    "Argument",
    "Option",
    "Parser",
]

import typing

if typing.TYPE_CHECKING:
    from hikari import messages
    from hikari import traits
    from hikari.api import event_dispatcher
    from hikari.api import shard as shard_
    from hikari.events import base_events

    from tanjun import errors


# To allow for stateless converters we accept both "Converter[...]" and "Type[StatelessConverter[...]]" where all the
# methods on "Type[StatelessConverter[...]]" need to be classmethods as it will not be initialised before calls are made
# to it.
ConverterT = typing.Union[
    typing.Callable[[str], typing.Any], "Converter[typing.Any]", "typing.Type[StatelessConverter[typing.Any]]"
]
"""Type hint of a converter used within a parser instance.

This may either be a callable which takes one position `builtins.string` argument,
a `Converter` instance or a `StatelessConverter` class.

`Converter` and `StatelessConverter` differ in the fact that
`StatelessConverter` is intended to be callable as a class where all it's methods
are class methods unlike `Converter` which will need to have initialised before
being registered as a listener.
"""
# TODO: be more specific about the structure of command functions using a callable protocol

CommandFunctionT = typing.Callable[..., typing.Coroutine[typing.Any, typing.Any, typing.Any]]
"""Type hint of the function a `Command` instance will operate on.

This will be called when executing a command and will need to take at least one
positional argument of type `Context` where any other required or optional
keyword or positional arguments will be based on the parser instance for the
command if applicable.

!!! note
    This will have to be asynchronous.
"""

LoadableT = typing.Callable[["Client"], None]
"""Type hint of the function used to load resources into a Tanjun client.

This should take one positional argument of type `Client` and return nothing.
This will be expected to initiate and resources like components to the client
through the use of it's protocol methods.
"""

ParserHookT = typing.Callable[
    ["Context", "errors.ParserError"], typing.Union[typing.Coroutine[typing.Any, typing.Any, None], None]
]
"""Type hint of the function used as a parser error hook.

This will be called whenever a `tanjun.errors.ParserError` is raised during the
command argument parsing stage, will have to take two positional arguments - of
type `Context` and `tanjun.errors.ParserError` - and may either be a
synchronous or asynchronous function which returns `builtins.None`
"""

ErrorHookT = typing.Callable[
    ["Context", BaseException], typing.Union[typing.Coroutine[typing.Any, typing.Any, None], None]
]
"""Type hint of the function used as a unexpected command error hook.

This will be called whenever a `builtins.BaseException` is raised during the
execution stage whenever the command function raises any exception except
`tanjun.errors.CommandError`,  will have to take two positional arguments - of
type `Context` and `builtins.BaseException` - and may either be a synchronous
or asynchronous function which returns `builtins.None`
"""

HookT = typing.Callable[["Context"], typing.Union[typing.Coroutine[typing.Any, typing.Any, None], None]]
"""Type hint of the function used as a general command hook.

This may be called during different stages of command execution (decided by
which hook this is registered as), will have to take one positional argument of
type `Context` and may be a synchronous or asynchronous function which returns
`builtins.None`.
"""

PreExecutionHookT = typing.Callable[["Context"], typing.Union[typing.Coroutine[typing.Any, typing.Any, bool], bool]]
"""Type hint of the function used as a pre-execution command hook.

This will be called before command function is executed, will have to take one
positional argument of type `Context` and may be a synchronous or asynchronous
function which returns `builtins.bool` (where returning `False` may cancel
execution of the current command).
"""

CheckT = typing.Callable[["Context"], typing.Union[bool, typing.Coroutine[typing.Any, typing.Any, bool]]]
"""Type hint of a general context check used with Tanjun `Executable` classes.

This may be registered with a `Executable` to add a rule which decides whether
it should execute for each context passed to it. This should take one positional
argument of type `Context` and may either be a synchronous or asynchronous
function which returns `builtins.bool` where returning `builtins.False` or
raising `tanjun.errors.FailedCheck` will indicate that the current context
shouldn't lead to an execution.
"""

ValueT = typing.TypeVar("ValueT", covariant=True)
"""A general type hint used for generic interfaces in Tanjun."""


@typing.runtime_checkable
class CommandDescriptor(typing.Protocol):
    """Descriptor of a command that's attached to a component."""

    __slots__: typing.Sequence[str] = ()

    @property
    def function(self) -> CommandFunctionT:
        """The function this Command will operate on.

        This will be called when executing a command and will need to take at
        least one positional argument of type `Context` where any other required
        or optional keyword or positional arguments will be based on `
        instance for the command if applicable.

        Returns
        -------
        CommandFunctionT
            The asynchronous function which should accept "self", and another
            positional argument of type `Context` along with the other accepted
            arguments being dependent on `CommandDescriptor.parser`.
        """
        raise NotImplementedError

    @property
    def is_owned(self) -> bool:
        """Whether this command descriptor belongs to a command group.

        Returns
        -------
        bool
            Whether this command descriptor belongs to a command group.
            If this is `builtins.True` then a component shouldn't try to load
            this into it's own commands.
        """
        raise NotImplementedError

    @property
    def metadata(self) -> typing.MutableMapping[typing.Any, typing.Any]:
        """Metadata which describes the command.

        Returns
        -------
        typing.MutableMapping[typing.Any, typing.Any]
            A mutable mapping of metadata which describes the command.
        """
        raise NotImplementedError

    @property
    def parser(self) -> typing.Optional[ParserDescriptor]:
        """The descriptor of this command's parser if set.

        Returns
        -------
        typing.Optional[ParserDescriptor]
            A descriptor of this command's parser if set, else `builtins.None`.
        """
        raise NotImplementedError

    @parser.setter
    def parser(self, parser: ParserDescriptor, /) -> None:
        raise NotImplementedError

    def add_check(self, check: CheckT, /) -> None:
        """Add a pre-execution check for this command descriptor.

        This will be run before execution to decide whether the command should
        be executed for the provided context.

        Parameters
        ----------
        check : CheckT
            The sync or async callable which takes one positional argument
            of type `Context` and return `builtins.bool` where returning
            `builtins.False` or raising `tanjun.errors.FailedCheck` will
            indicate that the current context shouldn't lead to an execution to
            register as a check.
        """
        raise NotImplementedError

    def with_check(self, check: CheckT, /) -> CheckT:
        """Decorator for adding a pre-execution check to this command.

        !!! note
            Unlike `CommandDescriptor.add_check`, this will return the passed
            check function to allow it to be used as a function decorator.

        Returns
        -------
        CheckT
            The passed check.
        """
        raise NotImplementedError

    def add_name(self, name: str, /) -> None:
        """Add a execution name to this command.

        This name is used to decide whether the command fits a given string or
        not.

        Parameters
        ----------
        name : str
            The name to add to this command.
        """
        raise NotImplementedError

    def build_command(self, component: Component, /) -> ExecutableCommand:
        """Build a command object from this descriptor.

        Parameters
        ----------
        component : ExecutableCommand
            The component this command is being built for.

        Returns
        -------
        ExecutableCommand
            The command object that was created.
        """
        raise NotImplementedError


@typing.runtime_checkable
class ListenerDescriptor(typing.Protocol):
    """Descriptor of a event listener that's attached to a component."""

    __slots__: typing.Sequence[str] = ()

    def build_listener(
        self, component: Component, /
    ) -> typing.Tuple[typing.Type[base_events.Event], event_dispatcher.CallbackT[typing.Any]]:
        """Build a listener from this descriptor.

        Parameters
        component : Component
            The component this listener is being built for.

        Returns
        -------
        typing.Tuple[typing.Type[hikari.events.base_events.Event], hikari.event_dispatcher.CallbackT[typing.Any]]
            A tuple of the event class this event listener should tbe registered
            for to the callable listener that should be registered.
        """
        raise NotImplementedError


class LoadableDescriptor:
    """Descriptor of a function used for loading a lib's resources into a Tanjun instance."""

    __slots__: typing.Sequence[str] = ()

    @property
    def load_function(self) -> LoadableT:
        """Function called to load these resources into a Tanjun client.

        Returns
        -------
        LoadableT
            The load function which should take one argument of type Client and
            return nothing. This should call methods on `Client` in-order to
            load it's pre-prepared resources.
        """
        raise NotImplementedError


@typing.runtime_checkable
class ParserDescriptor(typing.Protocol):
    """Descriptor of a parser for command descriptor."""

    __slots__: typing.Sequence[str] = ()

    @property
    def parameters(self) -> typing.Sequence[Parameter]:
        """Get the parameters rules set for this parser descriptor.

        Returns
        -------
        typing.Sequence[Parameter]
            A sequence of the parameters rules set for this parser descriptor.
        """
        raise NotImplementedError

    def add_parameter(self, parameter: Parameter, /) -> None:
        """Add a parameter to the parser's rules.

        !!! note
            For positional arguments this should add from left to right if the
            index isn't explicitly declared.

        Parameters
        ----------
        parameter : Parameter
            The parameter to add.
        """
        raise NotImplementedError

    def set_parameters(self, parameters: typing.Iterable[Parameter], /) -> None:
        """Set the parameters for this parser's rules.

        !!! note
            This will replace any previously set parameters.

        Parameters
        ----------
        parameters : typing.Iterable[Parameter]
            An iterable of the parameters to set for this parser.
        """
        raise NotImplementedError

    def build_parser(self, component: Component, /) -> Parser:
        """Build a parser object from this descriptor.

        Parameters
        ----------
        component : ExecutableCommand
            The component this command is being built for.

        Returns
        -------
        ExecutableCommand
            The command object that was created.
        """
        raise NotImplementedError


@typing.runtime_checkable
class Context(typing.Protocol):
    """Traits for the context of a command execution event."""

    __slots__: typing.Sequence[str] = ()

    @property
    def client(self) -> Client:
        """The Tanjun `Client` implementation this context was spawned by.

        Returns
        -------
        Client
            The Tanjun `Client` implementation this context was spawned by.
        """
        raise NotImplementedError

    @property
    def command(self) -> typing.Optional[ExecutableCommand]:
        raise NotImplementedError

    @command.setter
    def command(self, command: ExecutableCommand, /) -> None:
        raise NotImplementedError

    @property
    def content(self) -> str:
        raise NotImplementedError

    @content.setter
    def content(self, content: str, /) -> None:
        raise NotImplementedError

    @property
    def message(self) -> messages.Message:
        raise NotImplementedError

    @property
    def shard(self) -> shard_.GatewayShard:
        raise NotImplementedError

    @property
    def triggering_prefix(self) -> typing.Optional[str]:
        raise NotImplementedError

    @triggering_prefix.setter
    def triggering_prefix(self, triggering_prefix: str, /) -> None:
        raise NotImplementedError

    @property
    def triggering_name(self) -> typing.Optional[str]:
        raise NotImplementedError

    @triggering_name.setter
    def triggering_name(self, triggering_name: str, /) -> None:
        raise NotImplementedError


@typing.runtime_checkable
class Converter(typing.Protocol[ValueT]):
    __slots__: typing.Sequence[str] = ()

    async def convert(self, ctx: Context, argument: str, /) -> ValueT:
        raise NotImplementedError

    def bind_client(self, client: Client, /) -> None:
        raise NotImplementedError

    def bind_component(self, component: Component, /) -> None:
        raise NotImplementedError


@typing.runtime_checkable
class StatelessConverter(typing.Protocol[ValueT]):
    __slots__: typing.Sequence[str] = ()

    @classmethod
    async def convert(cls, ctx: Context, argument: str, /) -> ValueT:
        raise NotImplementedError

    @classmethod
    def bind_client(cls, client: Client, /) -> None:
        raise NotImplementedError

    def bind_component(self, component: Component, /) -> None:
        raise NotImplementedError


@typing.runtime_checkable
class Hooks(typing.Protocol):
    __slots__: typing.Sequence[str] = ()

    async def trigger_error(
        self, ctx: Context, /, exception: BaseException, *, hooks: typing.Optional[typing.AbstractSet[Hooks]] = None
    ) -> None:
        raise NotImplementedError

    async def trigger_parser_error(
        self, ctx: Context, /, exception: errors.ParserError, hooks: typing.Optional[typing.AbstractSet[Hooks]] = None,
    ) -> None:
        raise NotImplementedError

    async def trigger_post_execution(
        self, ctx: Context, /, *, hooks: typing.Optional[typing.AbstractSet[Hooks]] = None
    ) -> None:
        raise NotImplementedError

    async def trigger_pre_execution(
        self, ctx: Context, /, *, hooks: typing.Optional[typing.AbstractSet[Hooks]] = None,
    ) -> bool:
        raise NotImplementedError

    async def trigger_success(
        self, ctx: Context, /, *, hooks: typing.Optional[typing.AbstractSet[Hooks]] = None
    ) -> None:
        raise NotImplementedError


@typing.runtime_checkable
class Executable(typing.Protocol):
    __slots__: typing.Sequence[str] = ()

    @property
    def checks(self) -> typing.AbstractSet[CheckT]:
        raise NotImplementedError

    @property
    def hooks(self) -> typing.Optional[Hooks]:
        raise NotImplementedError

    @hooks.setter
    def hooks(self, hooks: typing.Optional[Hooks]) -> None:
        raise NotImplementedError

    def add_check(self, check: CheckT, /) -> None:
        raise NotImplementedError

    def remove_check(self, check: CheckT, /) -> None:
        raise NotImplementedError

    def with_check(self, check: CheckT, /) -> CheckT:
        raise NotImplementedError

    # As far as MYPY is concerned, unless you explicitly yield within an async function typed as returning an
    # AsyncIterator/AsyncGenerator you are returning an AsyncIterator/AsyncGenerator as the result of a coroutine.
    def check_context(self, ctx: Context, /, *, name_prefix: str = "") -> typing.AsyncIterator[FoundCommand]:
        raise NotImplementedError

    def check_name(self, name: str, /) -> typing.Iterator[FoundCommand]:
        raise NotImplementedError

    async def execute(self, ctx: Context, /, *, hooks: typing.Optional[typing.MutableSet[Hooks]] = None) -> bool:
        raise NotImplementedError


@typing.runtime_checkable
class FoundCommand(typing.Protocol):
    __slots__: typing.Sequence[str] = ()

    @property
    def command(self) -> ExecutableCommand:
        raise NotImplementedError

    @command.setter
    def command(self, command: ExecutableCommand, /) -> None:
        raise NotImplementedError

    @property
    def name(self) -> str:
        raise NotImplementedError

    @name.setter
    def name(self, name: typing.Optional[str], /) -> None:
        raise NotImplementedError


@typing.runtime_checkable
class ExecutableCommand(Executable, typing.Protocol):
    __slots__: typing.Sequence[str] = ()

    @property
    def component(self) -> typing.Optional[Component]:
        raise NotImplementedError

    @property
    def function(self) -> CommandFunctionT:
        raise NotImplementedError

    @property
    def metadata(self) -> typing.MutableMapping[typing.Any, typing.Any]:
        raise NotImplementedError

    @property
    def names(self) -> typing.AbstractSet[str]:
        raise NotImplementedError

    @property
    def parent(self) -> typing.Optional[ExecutableCommandGroup]:
        raise NotImplementedError

    @parent.setter
    def parent(self, parent: typing.Optional[ExecutableCommandGroup], /) -> None:
        raise NotImplementedError

    @property
    def parser(self) -> typing.Optional[Parser]:
        raise NotImplementedError

    @parser.setter
    def parser(self, parser: typing.Optional[Parser], /) -> None:
        raise NotImplementedError

    def add_name(self, name: str, /) -> None:
        raise NotImplementedError

    def remove_name(self, name: str, /) -> None:
        raise NotImplementedError

    def bind_client(self, client: Client, /) -> None:
        raise NotImplementedError

    def bind_component(self, component: Component, /) -> None:
        raise NotImplementedError


class ExecutableCommandGroup(ExecutableCommand, typing.Protocol):
    __slots__: typing.Sequence[str] = ()

    @property
    def commands(self) -> typing.AbstractSet[ExecutableCommand]:
        raise NotImplementedError

    def add_command(self, command: ExecutableCommand, /) -> None:
        raise NotImplementedError

    def remove_command(self, command: ExecutableCommand, /) -> None:
        raise NotImplementedError

    def with_command(
        self,
        name: str,
        /,
        *names: str,
        checks: typing.Optional[typing.Iterable[CheckT]] = None,
        hooks: typing.Optional[Hooks] = None,
        parser: typing.Optional[Parser] = None,
    ) -> typing.Callable[[CommandFunctionT], CommandFunctionT]:
        raise NotImplementedError


@typing.runtime_checkable
class Component(Executable, typing.Protocol):
    __slots__: typing.Sequence[str] = ()

    @property
    def client(self) -> typing.Optional[Client]:
        raise NotImplementedError

    @property
    def commands(self) -> typing.AbstractSet[ExecutableCommand]:
        raise NotImplementedError

    @property
    def listeners(
        self,
    ) -> typing.AbstractSet[typing.Tuple[typing.Type[base_events.Event], event_dispatcher.CallbackT[typing.Any]]]:
        raise NotImplementedError

    @property
    def metadata(self) -> typing.MutableMapping[typing.Any, typing.Any]:
        raise NotImplementedError

    def add_command(self, command: ExecutableCommand, /) -> None:
        raise NotImplementedError

    def remove_command(self, command: ExecutableCommand, /) -> None:
        raise NotImplementedError

    def add_listener(
        self, event: typing.Type[base_events.Event], listener: event_dispatcher.CallbackT[typing.Any], /
    ) -> None:
        raise NotImplementedError

    def remove_listener(
        self, event: typing.Type[base_events.Event], listener: event_dispatcher.CallbackT[typing.Any], /
    ) -> None:
        raise NotImplementedError

    def bind_client(self, client: Client, /) -> None:
        raise NotImplementedError

    async def close(self) -> None:
        raise NotImplementedError

    async def open(self) -> None:
        raise NotImplementedError


@typing.runtime_checkable
class Client(typing.Protocol):
    __slots__: typing.Sequence[str] = ()

    @property
    def cache_service(self) -> typing.Optional[traits.CacheAware]:
        raise NotImplementedError

    @property
    def components(self) -> typing.AbstractSet[Component]:
        raise NotImplementedError

    @property
    def dispatch_service(self) -> traits.DispatcherAware:
        raise NotImplementedError

    @property
    def hooks(self) -> typing.Optional[Hooks]:
        raise NotImplementedError

    @hooks.setter
    def hooks(self, hooks: typing.Optional[Hooks]) -> None:
        raise NotImplementedError

    @property
    def metadata(self) -> typing.MutableMapping[typing.Any, typing.Any]:
        raise NotImplementedError

    @property
    def prefixes(self) -> typing.AbstractSet[str]:
        raise NotImplementedError

    @property
    def rest_service(self) -> traits.RESTAware:
        raise NotImplementedError

    @property
    def shard_service(self) -> traits.ShardAware:
        raise NotImplementedError

    def add_component(self, component: Component, /) -> None:
        raise NotImplementedError

    def remove_component(self, component: Component, /) -> None:
        raise NotImplementedError

    def add_prefix(self, prefix: str, /) -> None:
        raise NotImplementedError

    def remove_prefix(self, prefix: str, /) -> None:
        raise NotImplementedError

    # As far as MYPY is concerned, unless you explicitly yield within an async function typed as returning an
    # AsyncIterator/AsyncGenerator you are returning an AsyncIterator/AsyncGenerator as the result of a coroutine.
    def check_context(self, ctx: Context, /) -> typing.AsyncIterator[FoundCommand]:
        raise NotImplementedError

    def check_name(self, name: str, /) -> typing.Iterator[FoundCommand]:
        raise NotImplementedError

    async def close(self) -> None:
        raise NotImplementedError

    async def open(self) -> None:
        raise NotImplementedError


class UndefinedDefault:
    __slots__: typing.Sequence[str] = ()


UNDEFINED_DEFAULT = UndefinedDefault()
"""A singleton used to represent no default for a parameter."""


@typing.runtime_checkable
class Parameter(typing.Protocol):
    __slots__: typing.Sequence[str] = ()

    @property
    def converters(self) -> typing.Optional[typing.Sequence[ConverterT]]:
        raise NotImplementedError

    @property
    def default(self) -> typing.Union[typing.Any, UndefinedDefault]:
        raise NotImplementedError

    @default.setter
    def default(self, default: typing.Union[typing.Any, UndefinedDefault], /) -> None:
        raise NotImplementedError

    @property
    def flags(self) -> typing.MutableMapping[str, typing.Any]:
        raise NotImplementedError

    @property
    def key(self) -> str:
        raise NotImplementedError

    @key.setter
    def key(self, key: str) -> None:
        raise NotImplementedError

    def add_converter(self, converter: ConverterT, /) -> None:
        raise NotImplementedError

    def remove_converter(self, converter: ConverterT, /) -> None:
        raise NotImplementedError

    def bind_client(self, client: Client, /) -> None:
        raise NotImplementedError

    def bind_component(self, component: Component, /) -> None:
        raise NotImplementedError

    async def convert(self, ctx: Context, value: str) -> typing.Any:
        raise NotImplementedError


@typing.runtime_checkable
class Argument(Parameter, typing.Protocol):
    __slots__: typing.Sequence[str] = ()


@typing.runtime_checkable
class Option(Parameter, typing.Protocol):
    __slots__: typing.Sequence[str] = ()

    @property
    def empty_value(self) -> typing.Union[typing.Any, UndefinedDefault]:
        raise NotImplementedError

    @empty_value.setter
    def empty_value(self, empty_value: typing.Union[typing.Any, UndefinedDefault], /) -> None:
        raise NotImplementedError

    @property
    def names(self) -> typing.Sequence[str]:
        raise NotImplementedError

    @names.setter
    def names(self, names: typing.Sequence[str], /) -> None:
        raise NotImplementedError


@typing.runtime_checkable
class Parser(typing.Protocol):
    __slots__: typing.Sequence[str] = ()

    @property
    def parameters(self) -> typing.Sequence[Parameter]:
        raise NotImplementedError

    def add_parameter(self, parameter: Parameter, /) -> None:
        raise NotImplementedError

    def remove_parameter(self, parameter: Parameter, /) -> None:
        raise NotImplementedError

    def set_parameters(self, parameters: typing.Iterable[Parameter], /) -> None:
        raise NotImplementedError

    def bind_client(self, client: Client, /) -> None:
        raise NotImplementedError

    def bind_component(self, component: Component, /) -> None:
        raise NotImplementedError

    async def parse(
        self, ctx: Context, /
    ) -> typing.Tuple[typing.Sequence[typing.Any], typing.Mapping[str, typing.Any]]:
        raise NotImplementedError



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

__all__: typing.Sequence[str] = ["as_loader", "Client"]

import asyncio
import importlib.util
import inspect
import itertools
import typing

from hikari import errors as hikari_errors
from hikari import traits as hikari_traits
from hikari.events import lifetime_events
from hikari.events import message_events
from yuyo import backoff

from tanjun import context
from tanjun import traits
from tanjun import utilities

if typing.TYPE_CHECKING:
    import pathlib
    import types


class _LoadableDescriptor(traits.LoadableDescriptor):
    def __init__(self, function: traits.LoadableT, /) -> None:
        self._function = function
        utilities.with_function_wrapping(self, "load_function")

    def __call__(self, client: traits.Client, /) -> None:
        self._function(client)

    @property
    def load_function(self) -> traits.LoadableT:
        return self._function


# This class is left unslotted as to allow it to "wrap" the underlying function
# by overwriting class attributes.
def as_loader(function: traits.LoadableT) -> traits.LoadableT:
    """Mark a function as being used to load Tanjun utilities from a module.

    Parameters
    ----------
    function : traits.LoadableT
        The function used to load Tanjun utilities from the a module. This
        should take one argument of type `tanjun.traits.Client`, return nothing
        and will be expected to initiate and add utilities such as components
        to the provided client using it's protocol methods.

    Returns
    -------
    traits.LoadableT
        The decorated load function.
    """
    return _LoadableDescriptor(function)


class Client(traits.Client):
    __slots__: typing.Sequence[str] = (
        "_cache",
        "_checks",
        "_components",
        "_dispatch",
        "_grab_mention_prefix",
        "hooks",
        "_metadata",
        "_prefixes",
        "_rest",
        "_shards",
    )

    def __init__(
        self,
        dispatch: hikari_traits.DispatcherAware,
        rest: typing.Optional[hikari_traits.RESTAware] = None,
        shard: typing.Optional[hikari_traits.ShardAware] = None,
        cache: typing.Optional[hikari_traits.CacheAware] = None,
        /,
        *,
        checks: typing.Optional[typing.Iterable[traits.CheckT]] = None,
        hooks: typing.Optional[traits.Hooks] = None,
        mention_prefix: bool = True,
        modules: typing.Optional[typing.Iterable[typing.Union[pathlib.Path, str]]] = None,
        prefixes: typing.Optional[typing.Iterable[str]] = None,
    ) -> None:
        if rest is not None:
            pass

        elif isinstance(cache, hikari_traits.RESTAware):
            rest = cache

        elif isinstance(dispatch, hikari_traits.RESTAware):
            rest = dispatch

        elif isinstance(shard, hikari_traits.RESTAware):
            rest = shard  # type: ignore[unreachable]

        else:
            raise ValueError("Missing RESTAware client implementation.")

        if shard is not None:
            pass

        elif isinstance(cache, hikari_traits.ShardAware):
            shard = cache

        elif isinstance(dispatch, hikari_traits.ShardAware):
            shard = dispatch

        elif isinstance(rest, hikari_traits.ShardAware):
            shard = rest

        else:
            raise ValueError("Missing ShardAware client implementation.")

        # Unlike `rest`, no provided Cache implementation just means this runs stateless.
        if cache is not None:
            pass

        elif isinstance(dispatch, hikari_traits.CacheAware):
            cache = dispatch

        elif isinstance(rest, hikari_traits.CacheAware):
            cache = rest

        elif isinstance(shard, hikari_traits.CacheAware):  # type: ignore[unreachable]
            cache = shard
        # TODO: logging or something to indicate this is running statelessly rather than statefully.

        self._checks: typing.MutableSet[traits.CheckT] = {self.check_human, *(checks or ())}
        self._cache = cache
        self._components: typing.MutableSet[traits.Component] = set()
        self._dispatch = dispatch
        self._grab_mention_prefix = mention_prefix
        self.hooks = hooks
        self._metadata: typing.MutableMapping[typing.Any, typing.Any] = {}
        self._prefixes = set(prefixes) if prefixes else set()
        self._rest = rest
        self._shards = shard
        self._dispatch.dispatcher.subscribe(lifetime_events.StartingEvent, self._on_starting_event)
        self._dispatch.dispatcher.subscribe(lifetime_events.StoppingEvent, self._on_stopping_event)

        if modules:
            self.load_from_modules(modules)

    async def __aenter__(self) -> Client:
        await self.open()
        return self

    async def __aexit__(
        self,
        exception_type: typing.Optional[typing.Type[BaseException]],
        exception: typing.Optional[BaseException],
        exception_traceback: typing.Optional[types.TracebackType],
    ) -> None:
        await self.close()

    def __repr__(self) -> str:
        return f"CommandClient <{type(self).__name__!r}, {len(self._components)} components, {self._prefixes}>"

    @property
    def cache_service(self) -> typing.Optional[hikari_traits.CacheAware]:
        return self._cache

    @property
    def checks(self) -> typing.AbstractSet[traits.CheckT]:
        return frozenset(self._checks)

    @property
    def components(self) -> typing.AbstractSet[traits.Component]:
        return frozenset(self._components)

    @property
    def dispatch_service(self) -> hikari_traits.DispatcherAware:
        return self._dispatch

    @property
    def metadata(self) -> typing.MutableMapping[typing.Any, typing.Any]:
        return self._metadata

    @property
    def prefixes(self) -> typing.AbstractSet[str]:
        return frozenset(self._prefixes)

    @property
    def rest_service(self) -> hikari_traits.RESTAware:
        return self._rest

    @property
    def shard_service(self) -> hikari_traits.ShardAware:
        return self._shards

    async def _on_starting_event(self, _: lifetime_events.StartingEvent, /) -> None:
        await self.open()

    async def _on_stopping_event(self, _: lifetime_events.StoppingEvent, /) -> None:
        await self.close()

    def add_check(self, check: traits.CheckT, /) -> None:
        self._checks.add(check)

    def remove_check(self, check: traits.CheckT, /) -> None:
        self._checks.remove(check)

    async def check(self, ctx: traits.Context, /) -> bool:
        return await utilities.gather_checks(utilities.await_if_async(check, ctx) for check in self._checks)

    def add_component(self, component: traits.Component, /) -> None:
        component.bind_client(self)
        self._components.add(component)

    def remove_component(self, component: traits.Component, /) -> None:
        self._components.remove(component)

    def add_prefix(self, prefix: str, /) -> None:
        self._prefixes.add(prefix)

    def remove_prefix(self, prefix: str, /) -> None:
        self._prefixes.remove(prefix)

    async def check_context(self, ctx: traits.Context, /) -> typing.AsyncIterator[traits.FoundCommand]:
        async for value in utilities.async_chain(component.check_context(ctx) for component in self._components):
            yield value

    @staticmethod
    def check_human(ctx: traits.Context, /) -> bool:
        return not ctx.message.author.is_bot and ctx.message.webhook_id is None

    def check_name(self, name: str, /) -> typing.Iterator[traits.FoundCommand]:
        yield from itertools.chain.from_iterable(component.check_name(name) for component in self._components)

    async def check_prefix(self, content: str, /) -> typing.Optional[str]:
        for prefix in self._prefixes:
            if content.startswith(prefix):
                return prefix

        return None

    async def close(self, *, deregister_listener: bool = True) -> None:
        self._dispatch.dispatcher.unsubscribe(message_events.MessageCreateEvent, self.on_message_create)

        if deregister_listener:
            await asyncio.gather(*(component.close() for component in self._components))

    async def open(self, *, register_listener: bool = True) -> None:
        await asyncio.gather(*(component.open() for component in self._components))

        if self._grab_mention_prefix:
            retry = backoff.Backoff(max_retries=4, maximum=30)

            async for _ in retry:
                try:
                    user = await self._rest.rest.fetch_my_user()
                    break

                except hikari_errors.RateLimitedError as exc:
                    if exc.retry_after > 30:
                        raise

                    retry.set_next_backoff(exc.retry_after)

                except hikari_errors.InternalServerError:
                    continue

            else:
                user = await self._rest.rest.fetch_my_user()

            self._grab_mention_prefix = False
            self._prefixes.add(f"<@{user.id}>")
            self._prefixes.add(f"<@!{user.id}>")

        if register_listener:
            self._dispatch.dispatcher.subscribe(message_events.MessageCreateEvent, self.on_message_create)

    def load_from_modules(self, modules: typing.Iterable[typing.Union[str, pathlib.Path]]) -> None:
        for module_path in modules:
            if isinstance(module_path, str):
                module = importlib.import_module(module_path)

            else:
                spec = importlib.util.spec_from_file_location(
                    module_path.name.rsplit(".", 1)[0], str(module_path.absolute())
                )
                module = importlib.util.module_from_spec(spec)

                if spec.loader is None:
                    raise RuntimeError(f"Invalid module provided {module_path}")

                # The type shedding is wrong here
                spec.loader.exec_module(module)  # type: ignore[attr-defined]

            for _, member in inspect.getmembers(module):
                if isinstance(member, traits.LoadableDescriptor):
                    member.load_function(self)

    async def on_message_create(self, event: message_events.MessageCreateEvent) -> None:
        if event.message.content is None:
            return

        if (prefix := await self.check_prefix(event.message.content)) is None:
            return

        content = event.message.content.lstrip()[len(prefix) :].lstrip()
        ctx = context.Context(self, content=content, message=event.message, triggering_prefix=prefix)

        if not await self.check(ctx):
            return

        hooks = {self.hooks,} if self.hooks else set()

        for component in self._components:
            if await component.execute(ctx, hooks=hooks):
                break



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""A collection of common standard checks designed for Tanjun commands."""

from __future__ import annotations

__all__: typing.Sequence[str] = [
    "with_dm_check",
    "with_guild_check",
    "with_nsfw_check",
    "with_sfw_check",
    "with_owner_check",
    "with_author_permission_check",
    "with_bot_permission_check",
]

import abc
import asyncio
import typing

from hikari import channels
from hikari import errors as hikari_errors
from hikari import guilds
from hikari import permissions as permissions_
from hikari.events import member_events
from yuyo import backoff

from tanjun import utilities

if typing.TYPE_CHECKING:
    from hikari import applications
    from hikari import snowflakes
    from hikari import traits as hikari_traits
    from hikari import users

    from tanjun import traits as tanjun_traits


CommandT = typing.TypeVar(
    "CommandT", "tanjun_traits.CommandDescriptor", "tanjun_traits.ExecutableCommand", contravariant=True
)


class ApplicationOwnerCheck:
    __slots__: typing.Sequence[str] = ("_application", "_client", "_fetch_task", "_lock")

    def __init__(self) -> None:
        self._application: typing.Optional[applications.Application] = None
        self._client: typing.Optional[tanjun_traits.Client] = None
        self._fetch_task: typing.Optional[asyncio.Task[None]] = None
        self._lock = asyncio.Lock()

    async def __call__(self, ctx: tanjun_traits.Context, /) -> bool:
        if ctx.client:
            self._client = ctx.client

        return await self.check(ctx)

    @staticmethod
    async def _fetch_application(rest: hikari_traits.RESTAware, /) -> applications.Application:  # type: ignore[return]
        retry = backoff.Backoff()
        async for _ in retry:
            try:
                return await rest.rest.fetch_application()

            except hikari_errors.RateLimitedError as exc:
                retry.set_next_backoff(exc.retry_after)

            except hikari_errors.InternalServerError:
                continue

    async def _fetch_loop(self) -> None:
        while True:
            # Update the application every 30 minutes.
            await asyncio.sleep(1_800)
            if self._client:
                try:
                    self._application = await self._fetch_application(self._client.rest_service)

                except hikari_errors.ForbiddenError:
                    pass

    async def _get_application(self, ctx: tanjun_traits.Context, /) -> applications.Application:
        if not self._application:
            async with self._lock:
                if self._application:
                    return self._application

                self._application = await self._fetch_application(ctx.client.rest_service)

                if not self._fetch_task:
                    self._fetch_task = asyncio.create_task(self._fetch_loop())

        return self._application

    async def check(self, ctx: tanjun_traits.Context, /) -> bool:
        application = await self._get_application(ctx)

        if not application.team and application.owner:
            return ctx.message.author.id == application.owner.id

        return bool(application.team and ctx.message.author.id in application.team.members)

    def close(self) -> None:
        if self._fetch_task:
            self._fetch_task.cancel()
            self._fetch_task = None

    async def open(self, client: tanjun_traits.Client, /) -> None:
        self.close()
        self._client = client
        self._fetch_task = asyncio.create_task(self._fetch_loop())

    async def update(self, *, rest: typing.Optional[hikari_traits.RESTAware] = None) -> None:
        if not rest and self._client:
            rest = self._client.rest_service

        elif not rest:
            raise ValueError("REST client must be provided when trying to update a closed application owner check.")

        self._application = await self._fetch_application(rest)


async def nsfw_check(ctx: tanjun_traits.Context, /) -> bool:
    channel: typing.Optional[channels.PartialChannel] = None
    if ctx.client.cache_service:
        channel = ctx.client.cache_service.cache.get_guild_channel(ctx.message.channel_id)

    if not channel:
        retry = backoff.Backoff(maximum=5, max_retries=4)
        channel = await utilities.fetch_resource(retry, ctx.message.fetch_channel)

    return channel.is_nsfw or False if isinstance(channel, channels.GuildChannel) else True


async def sfw_check(ctx: tanjun_traits.Context, /) -> bool:
    return not await nsfw_check(ctx)


def dm_check(ctx: tanjun_traits.Context, /) -> bool:
    return ctx.message.guild_id is None


def guild_check(ctx: tanjun_traits.Context, /) -> bool:
    return not dm_check(ctx)


class PermissionCheck(abc.ABC):
    __slots__: typing.Sequence[str] = ("permissions",)

    def __init__(self, permissions: typing.Union[permissions_.Permissions, int], /) -> None:
        self.permissions = permissions_.Permissions(permissions)

    async def __call__(self, ctx: tanjun_traits.Context, /) -> bool:
        return (self.permissions & await self.get_permissions(ctx)) == self.permissions

    @abc.abstractmethod
    async def get_permissions(self, ctx: tanjun_traits.Context, /) -> permissions_.Permissions:
        raise NotImplementedError


class AuthorPermissionCheck(PermissionCheck):
    __slots__: typing.Sequence[str] = ()

    async def get_permissions(self, ctx: tanjun_traits.Context, /) -> permissions_.Permissions:
        if not ctx.message.member:
            return utilities.ALL_PERMISSIONS

        return await utilities.calculate_permissions(ctx.client, ctx.message.member, channel=ctx.message.channel_id)


class BotPermissionsCheck(PermissionCheck):
    __slots__: typing.Sequence[str] = ("_lock", "_me" "_members")

    def __init__(self, permissions: typing.Union[permissions_.Permissions, int], /) -> None:
        super().__init__(permissions)
        self._lock = asyncio.Lock()
        self._me: typing.Optional[users.User] = None
        self._members: typing.MutableMapping[snowflakes.Snowflake, guilds.Member] = {}

    async def on_member_event(self, event: member_events.MemberEvent, /) -> None:
        if event.user_id != (await self._get_user(event.app, event.app)):
            return

        if isinstance(event, (member_events.MemberUpdateEvent, member_events.MemberCreateEvent)):
            self._members[event.user_id] = event.member

        elif event.user_id in self._members:
            del self._members[event.user_id]

    async def get_permissions(self, ctx: tanjun_traits.Context, /) -> permissions_.Permissions:
        if ctx.message.guild_id is None:
            return utilities.ALL_PERMISSIONS

        member = await self._get_member(ctx)
        return await utilities.calculate_permissions(ctx.client, member, channel=ctx.message.channel_id)

    async def _get_member(self, ctx: tanjun_traits.Context, /) -> guilds.Member:
        if ctx.message.guild_id is None:
            raise RuntimeError("Cannot get member for a DM Channel")

        if member := self._members.get(ctx.message.guild_id):
            return member

        user = await self._get_user(ctx.client.cache_service, ctx.client.rest_service)

        if ctx.client.cache_service and (
            member := ctx.client.cache_service.cache.get_member(ctx.message.guild_id, user.id)
        ):
            return member

        retry = backoff.Backoff(maximum=5, max_retries=4)
        return await utilities.fetch_resource(
            retry, ctx.client.rest_service.rest.fetch_member, ctx.message.guild_id, user.id,
        )

    async def _get_user(
        self, cache_service: typing.Optional[hikari_traits.CacheAware], rest_service: hikari_traits.RESTAware, /
    ) -> users.User:
        if not self._me:
            async with self._lock:
                if self._me:
                    return self._me

                if cache_service and (user := cache_service.cache.get_me()):
                    self._me = user

                else:
                    retry = backoff.Backoff(maximum=5, max_retries=4)
                    raw_user = await utilities.fetch_resource(retry, rest_service.rest.fetch_my_user)
                    self._me = raw_user

        return self._me


def with_dm_check(command: CommandT, /) -> CommandT:
    """Only let a command run in a DM channel.

    Parameters
    ----------
    command : CommandT
        The command to add this check to.

    Returns
    -------
    CommandT
        The command this check was added to.
    """
    command.add_check(guild_check)
    return command


def with_guild_check(command: CommandT, /) -> CommandT:
    """Only let a command run in a guild channel.

    Parameters
    ----------
    command : CommandT
        The command to add this check to.

    Returns
    -------
    CommandT
        The command this check was added to.
    """
    command.add_check(dm_check)
    return command


def with_nsfw_check(command: CommandT, /) -> CommandT:
    """Only let a command run in a channel that's marked as nsfw.

    Parameters
    ----------
    command : CommandT
        The command to add this check to.

    Returns
    -------
    CommandT
        The command this check was added to.
    """
    command.add_check(nsfw_check)
    return command


def with_sfw_check(command: CommandT, /) -> CommandT:
    """Only let a command run in a channel that's not marked as nsfw.

    Parameters
    ----------
    command : CommandT
        The command to add this check to.

    Returns
    -------
    CommandT
        The command this check was added to.
    """
    command.add_check(sfw_check)
    return command


def with_owner_check(command: CommandT, /) -> CommandT:
    """Only let a command run if it's being triggered by one of the bot's owners.

    !!! note
        This is based on the owner(s) of the bot's application and will account
        for team owners as well.

    Parameters
    ----------
    command : CommandT
        The command to add this check to.

    Returns
    -------
    CommandT
        The command this check was added to.
    """
    command.add_check(ApplicationOwnerCheck())
    return command


def with_author_permission_check(
    permissions: typing.Union[permissions_.Permissions, int], /
) -> typing.Callable[[CommandT], CommandT]:
    """Only let a command run if the author has certain permissions in the current channel.

    !!! note
        This will always pass for commands triggered in DM channels.

    Parameters
    ----------
    permissions: typing.Union[hikari.permissions.Permissions, builtins.int]
        The permission(s) required for this command to run.

    Returns
    -------
    typing.Callable[[CommandT], CommandT]
        A command decorator function which adds the check.
    """

    def decorator(command: CommandT, /) -> CommandT:
        command.add_check(AuthorPermissionCheck(permissions))
        return command

    return decorator


def with_bot_permission_check(
    permissions: typing.Union[permissions_.Permissions, int], /
) -> typing.Callable[[CommandT], CommandT]:
    """Only let a command run if we have certain permissions in the current channel.

    !!! note
        This will always pass for commands triggered in DM channels.

    Parameters
    ----------
    permissions: typing.Union[hikari.permissions.Permissions, builtins.int]
        The permission(s) required for this command to run.

    Returns
    -------
    typing.Callable[[CommandT], CommandT]
        A command decorator function which adds the check.
    """

    def decorator(command: CommandT, /) -> CommandT:
        command.add_check(BotPermissionsCheck(permissions))
        return command

    return decorator



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""The errors and warnings raised within and by Tanjun."""
from __future__ import annotations

__all__: typing.Sequence[str] = [
    "CommandError",
    "ConversionError",
    "FailedCheck",
    "NotEnoughArgumentsError",
    "TooManyArgumentsError",
    "ParserError",
    "TanjunError",
    "TanjunWarning",
    "StateWarning",
]

import typing

from tanjun import traits


class TanjunError(Exception):
    """The base class for all errors raised by Tanjun."""

    __slots__: typing.Sequence[str] = ()


class TanjunWarning(RuntimeWarning):
    """The base class for all warnings raised by Tanjun."""

    __slots__: typing.Sequence[str] = ()


class CommandError(TanjunError):
    """Error raised to end command execution.

    Parameters
    ----------
    message : typing.Optional[str]
        If this is a non-empty string then this message should will be sent as
        a response to the message that triggered the current command otherwise
        `builtins.None` or `""` will silently end command execution.
    """

    __slots__: typing.Sequence[str] = ("message",)

    # None or empty string == no response
    message: typing.Optional[str]
    """The response error message.

    If this is an empty string or `builtins.None` then this will silently end
    command execution otherwise Tanjun will try to send the string message in
    response.
    """

    def __init__(self, message: typing.Optional[str], /) -> None:
        if message and len(message) > 2000:
            raise ValueError("Error message cannot be over 2_000 characters long.")

        self.message = message

    def __repr__(self) -> str:
        return f"{type(self).__name__} <{self.message}>"

    def __str__(self) -> str:
        return self.message or ""


class FailedCheck(TanjunError, RuntimeError):
    """Error raised as an alternative to returning `False` in a check."""

    __slots__: typing.Sequence[str] = ()


class ParserError(TanjunError, ValueError):
    """Base error raised by a parser or parameter during parsing.

    !!! note
        Other error raised by the parser should subclass this error.

    Parameters
    ----------
    message : str
        String message for this error.
    parameter : typing.Optional[traits.Parameter]
        The parameter which caused this error, should be `builtins.None` if not
        applicable.
    """

    __slots__: typing.Sequence[str] = ("message", "parameter")

    message: str
    """String message for this error.

    !!! note
        This may be used as a command response message.
    """

    parameter: typing.Optional[traits.Parameter]
    """Parameter this was raised for.

    !!! note
        This will be `builtin.None` if it was raised while parsing the provided
        message content.
    """

    def __init__(self, message: str, parameter: typing.Optional[traits.Parameter], /) -> None:
        self.message = message
        self.parameter = parameter

    def __str__(self) -> str:
        return self.message


class ConversionError(ParserError):
    """Error raised by a parser parameter when it failed to converter a value.

    Parameters
    ----------
    parameter : tanjun.traits.Parameter
        The parameter this was raised by.
    errors : typing.Iterable[ValueError]
        An iterable of the source value errors which were raised during conversion/
    """

    __slots__: typing.Sequence[str] = ("errors",)

    errors: typing.Sequence[ValueError]
    """Sequence of the errors that were caught during conversion for this parameter."""

    parameter: traits.Parameter
    """Parameter this error was raised for."""

    def __init__(self, parameter: traits.Parameter, errors: typing.Iterable[ValueError], /) -> None:
        option_or_argument = "option" if isinstance(parameter, traits.Option) else "argument"
        super().__init__(f"Couldn't convert {option_or_argument} '{parameter.key}'", parameter)
        self.errors = tuple(errors)


class NotEnoughArgumentsError(ParserError):
    """Error raised by the parser when not enough arguments are found for a parameter.

    Parameters
    ----------
    parameter : tanjun.traits.Parameter
        The parameter this error was raised for
    """

    __slots__: typing.Sequence[str] = ()

    parameter: traits.Parameter
    """Parameter this error was raised for."""

    def __init__(self, message: str, parameter: traits.Parameter, /) -> None:
        super().__init__(message, parameter)


class TooManyArgumentsError(ParserError):
    """Error raised by the parser when too many arguments are found for a parameter.

    Parameters
    ----------
    parameter : tanjun.traits.Parameter
        The parameter this error was raised for
    """

    __slots__: typing.Sequence[str] = ()

    parameter: traits.Parameter
    """Parameter this error was raised for."""

    def __init__(self, message: str, parameter: traits.Parameter, /) -> None:
        super().__init__(message, parameter)


class StateWarning(RuntimeWarning):
    """Warning raised when a utility is loaded without access to state stores it depends on."""

    __slots__: typing.Sequence[str] = ()



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

__all__: typing.Sequence[str] = ["as_command", "as_event", "as_group", "Component"]

import copy
import inspect
import itertools
import types
import typing

from hikari.events import base_events

from tanjun import commands
from tanjun import traits
from tanjun import utilities

if typing.TYPE_CHECKING:
    from hikari.api import event_dispatcher


# This class is left unslotted as to allow it to "wrap" the underlying function
# by overwriting class attributes.
class CommandDescriptor(traits.CommandDescriptor):
    def __init__(
        self,
        checks: typing.Optional[typing.Iterable[traits.CheckT]],
        function: traits.CommandFunctionT,
        hooks: typing.Optional[traits.Hooks],
        names: typing.Sequence[str],
        parser: typing.Optional[traits.ParserDescriptor],
    ) -> None:
        self._checks = list(checks) if checks else []
        self._function = function
        self._hooks = hooks
        self._metadata: typing.MutableMapping[typing.Any, typing.Any] = {}
        self._names = list(names)
        self._parser = parser
        utilities.with_function_wrapping(self, "function")

    async def __call__(self, ctx: traits.Context, *args: typing.Any, **kwargs: typing.Any) -> typing.Any:
        return await self._function(ctx, *args, **kwargs)

    def __repr__(self) -> str:
        return f"CommandDescriptor <{self._function, self._names}>"

    @property
    def function(self) -> traits.CommandFunctionT:
        return self._function

    @property
    def is_owned(self) -> bool:
        return False

    @property
    def metadata(self) -> typing.MutableMapping[typing.Any, typing.Any]:
        return self._metadata

    @property
    def parser(self) -> typing.Optional[traits.ParserDescriptor]:
        return self._parser

    @parser.setter
    def parser(self, parser_: typing.Optional[traits.ParserDescriptor]) -> None:
        self._parser = parser_

    def add_check(self, check: traits.CheckT, /) -> None:
        self._checks.append(check)

    def with_check(self, check: traits.CheckT, /) -> traits.CheckT:
        self.add_check(check)
        return check

    def add_name(self, name: str, /) -> None:
        self._names.append(name)

    def build_command(self, component: traits.Component, /) -> traits.ExecutableCommand:
        command = commands.Command(
            types.MethodType(self._function, component),
            *self._names,
            checks=self._checks.copy(),
            hooks=copy.copy(self._hooks),
            metadata=dict(self._metadata),
            parser=self.parser.build_parser(component) if self.parser else None,
        )
        command.bind_component(component)
        return command


# This class is left unslotted as to allow it to "wrap" the underlying function
# by overwriting class attributes.
class _OwnedCommandDescriptor(CommandDescriptor):
    def is_owned(self) -> bool:
        return True


# This class is left unslotted as to allow it to "wrap" the underlying function
# by overwriting class attributes.
class CommandGroupDescriptor(CommandDescriptor):
    def __init__(
        self,
        checks: typing.Optional[typing.Iterable[traits.CheckT]],
        function: traits.CommandFunctionT,
        hooks: typing.Optional[traits.Hooks],
        names: typing.Sequence[str],
        parser: typing.Optional[traits.ParserDescriptor],
    ) -> None:
        super().__init__(checks, function, hooks, names, parser)
        self._commands: typing.MutableSequence[CommandDescriptor] = []

    def __repr__(self) -> str:
        return f"CommandGroupDescriptor <{self._function, self._names}, commands: {len(self._commands)}>"

    def build_command(self, component: traits.Component, /) -> traits.ExecutableCommandGroup:
        group = commands.CommandGroup(
            types.MethodType(self._function, component),
            *self._names,
            checks=self._checks.copy(),
            hooks=copy.copy(self._hooks),
            metadata=dict(self._metadata),
            parser=self.parser.build_parser(component) if self.parser else None,
        )
        group.bind_component(component)

        for descriptor in self._commands:
            group.add_command(descriptor.build_command(component))

        return group

    def with_command(
        self,
        name: str,
        /,
        *names: str,
        checks: typing.Optional[typing.Iterable[traits.CheckT]] = None,
        hooks: typing.Optional[traits.Hooks] = None,
        parser: typing.Optional[traits.ParserDescriptor] = None,
    ) -> typing.Callable[[traits.CommandFunctionT], CommandDescriptor]:
        def decorator(function: traits.CommandFunctionT) -> CommandDescriptor:
            descriptor = _OwnedCommandDescriptor(checks, function, hooks, (name, *names), parser)
            self._commands.append(descriptor)
            return descriptor

        return decorator


# This class is left unslotted as to allow it to "wrap" the underlying function
# by overwriting class attributes.
class ListenerDescriptor(traits.ListenerDescriptor):
    def __init__(
        self, event: typing.Type[base_events.Event], function: event_dispatcher.CallbackT[typing.Any], /
    ) -> None:
        self.event = event
        self.listener = function
        utilities.with_function_wrapping(self, "listener")

    async def __call__(self, event: base_events.Event, /) -> None:
        return await self.listener(event)

    def __repr__(self) -> str:
        return f"ListenerDescriptor for {self.event.__name__}: {self.listener}>"

    def build_listener(
        self, component: traits.Component, /
    ) -> typing.Tuple[typing.Type[base_events.Event], event_dispatcher.CallbackT[typing.Any]]:
        return self.event, types.MethodType(self.listener, component)


def as_command(
    name: str,
    /,
    *names: str,
    checks: typing.Optional[typing.Iterable[traits.CheckT]] = None,
    hooks: typing.Optional[traits.Hooks] = None,
    parser: typing.Optional[traits.ParserDescriptor] = None,
) -> typing.Callable[[traits.CommandFunctionT], CommandDescriptor]:
    """Declare a command descriptor on a component's class.

    The returned descriptor will be loaded into the component it's attached to
    during initialisation.

    Parameters
    ----------
    name : str
        The name for this command.

    Other Parameters
    ----------------
    *names : str
        Additional names for this command passed as variable positional arguments.
    checks : typing.Optional[typing.Iterable[traits.CheckT]]
        An iterable of async or non-async check functions which should take
        one positional argument of type `Context` and return `builtins.bool`
        or raise `tanjun.errors.FailedCheck` where `FailedCheck` or `False` will
        prevent the command from being chosen for execution.
    hooks : typing.Optional[traits.Hooks]
        The execution hooks this command should be calling.
    parser : typing.Optional[traits.ParserDescriptor]
        A descriptor of the parser this command should use.

    Returns
    -------
    typing.Callable[[traits.CommandFunctionT], CommandDescriptor]
        A decorator function for a command function.

    Examples
    --------
    Methods and attributes on the returned descriptor can be modified directly
    or using decorators in-order to fine tune the behaviour of this command
    (e.g. set a parser for the command) as shown below.

    ```python
    import tanjun

    class MyComponent(tanjun.components.Component):
        @tanjun.parsing.with_greedy_argument("content", converters=None)
        @tanjun.parsing.with_parser
        @tanjun.checks.with_owner_check
        @tanjun.components.as_command("echo")
        async def echo_command(self, ctx: tanjun.traits.Context, /, content: str) -> None:
            await ctx.message.reply(content)
    ```
    """

    def decorator(function: traits.CommandFunctionT, /) -> CommandDescriptor:
        return CommandDescriptor(checks, function, hooks, (name, *names), parser)

    return decorator


def as_group(
    name: str,
    /,
    *names: str,
    checks: typing.Optional[typing.Iterable[traits.CheckT]] = None,
    hooks: typing.Optional[traits.Hooks] = None,
    parser: typing.Optional[traits.ParserDescriptor] = None,
) -> typing.Callable[[traits.CommandFunctionT], CommandGroupDescriptor]:
    """Declare a command group descriptor on a component's class.

    The returned descriptor will be loaded into the component it's attached to
    during initialisation.

    Parameters
    ----------
    name : str
        The name for this command group.

    Other Parameters
    ----------------
    *names : str
        Additional names for this command group passed as variable positional arguments.
    checks : typing.Optional[typing.Iterable[traits.CheckT]]
        An iterable of async or non-async check functions which should take
        one positional argument of type `Context` and return `builtins.bool`
        or raise `tanjun.errors.FailedCheck` where `FailedCheck` or `False` will
        prevent the command from being chosen for execution.
    hooks : typing.Optional[traits.Hooks]
        The execution hooks this command should be calling.
    parser : typing.Optional[traits.ParserDescriptor]
        A descriptor of the parser this command should use.

    Returns
    -------
    typing.Callable[[traits.CommandFunctionT], CommandGroupDescriptor]
        A decorator function for a command function.

    Examples
    --------
    Methods and attributes on the returned descriptor can be modified directly
    or using decorators in-order to fine tune the behaviour of this command
    (e.g. set a parser for the command) as shown below.

    ```python
    import tanjun

    class MyComponent(tanjun.components.Component):
        @tanjun.parsing.with_greedy_argument("content", converters=None)
        @tanjun.parsing.with_parser
        @tanjun.checks.with_owner_check
        @tanjun.components.as_group("help")
        async def help(self, ctx: tanjun.traits.Context, /, content: str) -> None:
            await ctx.message.reply(f"`{content}` is not a valid help command")
    ```

    The `CommandGroupDescriptor.with_command` can be called on the returned
    descriptor in-order to add sub-commands to this group as shown below.

    ```python
    import tanjun

    class MyComponent(tanjun.components.Component):
        @tanjun.parsing.with_greedy_argument("content", converters=None)
        @tanjun.parsing.with_parser
        @tanjun.components.group("help")
        async def help(self, ctx: tanjun.traits.Context, /, content: str) -> None:
            await ctx.message.reply(f"`{content}` is not a valid help command")

        @help.with_command("me")
        async def help_me(self, ctx: tanjun.traits.Context) -> None:
            await ctx.message.reply(f"There is no reset for you, {ctx.message.author}")
    ```
    """

    def decorator(function: traits.CommandFunctionT, /) -> CommandGroupDescriptor:
        return CommandGroupDescriptor(checks, function, hooks, (name, *names), parser)

    return decorator


EventDecoratorT = typing.Callable[
    ["event_dispatcher.CallbackT[event_dispatcher.EventT_inv]"],
    "event_dispatcher.CallbackT[event_dispatcher.EventT_inv]",
]


def as_event(cls: typing.Type[event_dispatcher.EventT_inv], /) -> EventDecoratorT[event_dispatcher.EventT_inv]:
    """Declare an event listener

    The returned descriptor will be loaded into the component it's attached to
    during initialisation.

    Parameters
    ----------
    cls : typing.Type[hikari.api.event_dispatcher.EventT_inv]
        The Hikari event class this listener should be called for.

    Returns
    -------
    EventDecoratorT[event_dispatcher.EventT_inv]
        The callable event decorator which takes the event's asynchronous
        listener function (with signature {self, event, /}) as it's only
        positional argument and returns a listener descriptor.

    Examples
    --------
    ```python
    from hikari import events
    import tanjun

    class MyComponent(tanjun.components.Component):
        @tanjun.components.event(events.GuildVisibilityEvent):
        async def on_guild_visibility_event(self, event: events.GuildVisibilityEvent) -> None:
            ...
    ```
    """

    def decorator(
        function: event_dispatcher.CallbackT[event_dispatcher.EventT_inv], /
    ) -> event_dispatcher.CallbackT[event_dispatcher.EventT_inv]:
        return ListenerDescriptor(cls, function)

    return decorator


# This class isn't slotted to let us overwrite command and event methods during initialisation by making sure
# class properties aren't read only
class Component(traits.Component):
    started: bool
    """Whether this component has been "started" yet.

    When this is `builtins.False` executing the cluster will always do nothing.
    """

    def __init__(
        self,
        *,
        checks: typing.Optional[typing.Iterable[traits.CheckT]] = None,
        hooks: typing.Optional[traits.Hooks] = None,
    ) -> None:
        self._checks = set(checks) if checks else set()
        self._client: typing.Optional[traits.Client] = None
        self._commands: typing.MutableSet[traits.ExecutableCommand] = set()
        self._hooks = hooks
        self._listeners: typing.MutableSet[
            typing.Tuple[typing.Type[base_events.Event], event_dispatcher.CallbackT[typing.Any]]
        ] = set()
        self._metadata: typing.MutableMapping[typing.Any, typing.Any] = {}
        self.started = False
        self._load_from_properties()

    def __repr__(self) -> str:
        return f"Component <{type(self).__name__}, {len(self._commands)} commands>"

    @property
    def checks(self) -> typing.AbstractSet[traits.CheckT]:
        return frozenset(self._checks)

    @property
    def client(self) -> typing.Optional[traits.Client]:
        return self._client

    @property
    def commands(self) -> typing.AbstractSet[traits.ExecutableCommand]:
        return frozenset(self._commands)

    @property
    def hooks(self) -> typing.Optional[traits.Hooks]:
        return self._hooks

    # Seeing as this class isn't slotted we cannot overwrite settable properties with instance variables.
    @hooks.setter
    def hooks(self, hooks_: traits.Hooks, /) -> None:
        self._hooks = hooks_

    @property
    def listeners(
        self,
    ) -> typing.AbstractSet[typing.Tuple[typing.Type[base_events.Event], event_dispatcher.CallbackT[typing.Any]]]:
        return frozenset(self._listeners)

    @property
    def metadata(self) -> typing.MutableMapping[typing.Any, typing.Any]:
        return self._metadata

    def add_check(self, check: traits.CheckT, /) -> None:
        self._checks.add(check)

    def remove_check(self, check: traits.CheckT, /) -> None:
        self._checks.remove(check)

    def with_check(self, check: traits.CheckT, /) -> traits.CheckT:
        self.add_check(check)
        return check

    def add_command(self, command: typing.Union[traits.ExecutableCommand, traits.CommandDescriptor], /) -> None:
        command = command.build_command(self) if isinstance(command, traits.CommandDescriptor) else command
        self._commands.add(command)

    def remove_command(self, command: traits.ExecutableCommand, /) -> None:
        self._commands.remove(command)

    def add_listener(
        self, event: typing.Type[base_events.Event], listener: event_dispatcher.CallbackT[typing.Any], /
    ) -> None:
        self._listeners.add((event, listener))

        if self.started and self._client:
            self._client.dispatch_service.dispatcher.subscribe(event, listener)

    def remove_listener(
        self, event: typing.Type[base_events.Event], listener: event_dispatcher.CallbackT[typing.Any], /
    ) -> None:
        self._listeners.remove((event, listener))

        if self.started and self._client:
            self._client.dispatch_service.dispatcher.unsubscribe(event, listener)

    def bind_client(self, client: traits.Client, /) -> None:
        self._client = client
        for event_, listener in self._listeners:
            self._client.dispatch_service.dispatcher.subscribe(event_, listener)

        for command_ in self._commands:
            command_.bind_client(client)

    async def check_context(
        self, ctx: traits.Context, /, *, name_prefix: str = ""
    ) -> typing.AsyncIterator[traits.FoundCommand]:
        if await utilities.gather_checks(utilities.await_if_async(check, ctx) for check in self._checks):
            async for value in utilities.async_chain(
                command_.check_context(ctx, name_prefix=name_prefix) for command_ in self._commands
            ):
                yield value

    def check_name(self, name: str, /) -> typing.Iterator[traits.FoundCommand]:
        yield from itertools.chain.from_iterable(command_.check_name(name) for command_ in self._commands)

    async def close(self) -> None:
        if not self.started:
            return

        self.started = False
        if self._client:
            for event_, listener in self._listeners:
                self._client.dispatch_service.dispatcher.unsubscribe(event_, listener)

    async def open(self) -> None:
        if self.started:
            return

        # This is duplicated between both open and bind_cluster to ensure that these are registered
        # as soon as possible the first time this is binded to a client and that these are
        # re-registered everytime an object is restarted.
        if self._client:
            for event_, listener in self._listeners:
                try:
                    self._client.dispatch_service.dispatcher.unsubscribe(event_, listener)
                except (KeyError, ValueError, LookupError):  # TODO: what does hikari raise?
                    continue

                self._client.dispatch_service.dispatcher.subscribe(event_, listener)

        self.started = True

    async def execute(
        self, ctx: traits.Context, /, *, hooks: typing.Optional[typing.MutableSet[traits.Hooks]] = None
    ) -> bool:
        if not self.started:
            return False

        async for command_ in self.check_context(ctx):
            ctx.triggering_name = command_.name
            ctx.content = ctx.content[len(command_.name) :].lstrip()
            # Only add our hooks and set command if we're sure we'll be executing the command here.
            ctx.command = command_.command

            if self.hooks and hooks:
                hooks.add(self.hooks)

            elif self.hooks:
                hooks = {self.hooks}

            await command_.command.execute(ctx, hooks=hooks)
            return True

        return False

    def _load_from_properties(self) -> None:
        for name, member in inspect.getmembers(self):
            if isinstance(member, traits.CommandDescriptor):
                result = member.build_command(self)

                # We don't want to load in commands which belong to a command group here.
                if not member.is_owned:
                    self.add_command(result)

                setattr(self, name, result.function)

            elif isinstance(member, traits.ListenerDescriptor):
                event_, listener = member.build_listener(self)
                self.add_listener(event_, listener)
                setattr(self, name, listener)



# -*- coding: utf-8 -*-
# cython: language_level=3
# BSD 3-Clause License
#
# Copyright (c) 2020, Faster Speeding
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
from __future__ import annotations

__all__: typing.Sequence[str] = ["Hooks"]

import asyncio
import typing

from tanjun import errors
from tanjun import traits
from tanjun import utilities
from tanjun.traits import Context


class Hooks(traits.Hooks):
    __slots__: typing.Sequence[str] = ("_error", "_parser_error", "_pre_execution", "_post_execution", "_success")

    def __init__(
        self,
        *,
        on_error: typing.Optional[traits.ErrorHookT] = None,
        parser_error: typing.Optional[traits.ParserHookT] = None,
        pre_execution: typing.Optional[traits.PreExecutionHookT] = None,
        post_execution: typing.Optional[traits.HookT] = None,
        on_success: typing.Optional[traits.HookT] = None,
    ) -> None:
        self._error = on_error
        self._parser_error = parser_error
        self._pre_execution = pre_execution
        self._post_execution = post_execution
        self._success = on_success

    def __repr__(self) -> str:
        return (
            f"Hooks <{self._error!r}, {self._parser_error!r}, {self._pre_execution!r}, "
            f"{self._post_execution!r}, {self._success!r}>"
        )

    def with_on_error(self, hook: typing.Optional[traits.ErrorHookT], /) -> typing.Optional[traits.ErrorHookT]:
        self._error = hook
        return hook

    def with_on_parser_error(self, hook: typing.Optional[traits.ParserHookT], /) -> typing.Optional[traits.ParserHookT]:
        self._parser_error = hook
        return hook

    def with_post_execution(self, hook: typing.Optional[traits.HookT], /) -> typing.Optional[traits.HookT]:
        self._post_execution = hook
        return hook

    def with_pre_execution(
        self, hook: typing.Optional[traits.PreExecutionHookT], /
    ) -> typing.Optional[traits.PreExecutionHookT]:
        self._pre_execution = hook
        return hook

    def with_on_success(self, hook: typing.Optional[traits.HookT], /) -> typing.Optional[traits.HookT]:
        self._success = hook
        return hook

    async def trigger_error(
        self,
        ctx: Context,
        /,
        exception: BaseException,
        *,
        hooks: typing.Optional[typing.AbstractSet[traits.Hooks]] = None,
    ) -> None:  # TODO: return True to indicate "raise" else False or None to suppress
        if self._error:
            await utilities.await_if_async(self._error, ctx, exception)

        if hooks:
            await asyncio.gather(*(hook.trigger_error(ctx, exception) for hook in hooks))

    async def trigger_parser_error(
        self,
        ctx: Context,
        /,
        exception: errors.ParserError,
        hooks: typing.Optional[typing.AbstractSet[traits.Hooks]] = None,
    ) -> None:
        if self._parser_error:
            await utilities.await_if_async(self._parser_error, ctx, exception)

        if hooks:
            await asyncio.gather(*(hook.trigger_parser_error(ctx, exception) for hook in hooks))

    async def trigger_post_execution(
        self, ctx: Context, /, *, hooks: typing.Optional[typing.AbstractSet[traits.Hooks]] = None
    ) -> None:
        if self._post_execution:
            await utilities.await_if_async(self._post_execution, ctx)

        if hooks:
            await asyncio.gather(*(hook.trigger_post_execution(ctx) for hook in hooks))

    async def trigger_pre_execution(
        self, ctx: Context, /, *, hooks: typing.Optional[typing.AbstractSet[traits.Hooks]] = None,
    ) -> bool:
        if self._pre_execution and await utilities.await_if_async(self._pre_execution, ctx) is False:
            return False

        if hooks:
            return await utilities.gather_checks(hook.trigger_pre_execution(ctx) for hook in hooks)

        return True

    async def trigger_success(
        self, ctx: Context, /, *, hooks: typing.Optional[typing.AbstractSet[traits.Hooks]] = None
    ) -> None:
        if self._success:
            await utilities.await_if_async(self._success, ctx)

        if hooks:
            await asyncio.gather(*(hook.trigger_success(ctx) for hook in hooks))
